CREATE OR REPLACE PROCEDURE test_cursor_performance (approach IN VARCHAR2)
IS
   CURSOR cur
   IS
      SELECT *
        FROM all_source
       WHERE ROWNUM < 100001;

   one_row       cur%ROWTYPE;

   TYPE t IS TABLE OF cur%ROWTYPE
      INDEX BY PLS_INTEGER;

   many_rows     t;

   last_timing   NUMBER;

   cntr number := 0;

   PROCEDURE start_timer
   IS
   BEGIN
      last_timing := DBMS_UTILITY.get_cpu_time;
   END;

   PROCEDURE show_elapsed_time (message_in IN VARCHAR2 := NULL)
   IS
   BEGIN
      DBMS_OUTPUT.put_line (
            '"'
         || message_in
         || '" completed in: '
         || TO_CHAR (
               ROUND ( (DBMS_UTILITY.get_cpu_time - last_timing) / 100, 2)));
   END;
BEGIN
   start_timer;

   CASE approach
      WHEN 'implicit cursor for loop'
      THEN
         FOR j IN cur
         LOOP
            cntr := cntr + 1;
         END LOOP;

         DBMS_OUTPUT.put_line (cntr);

      WHEN 'explicit open, fetch, close'
      THEN
         OPEN cur;

         LOOP
            FETCH cur INTO one_row;

            cntr := cntr + 1;

            EXIT WHEN cur%NOTFOUND;
         END LOOP;

         DBMS_OUTPUT.put_line (cntr);

         CLOSE cur;
      WHEN 'bulk fetch'
      THEN
         OPEN cur;

         LOOP
            FETCH cur BULK COLLECT INTO many_rows LIMIT 100;

            EXIT WHEN many_rows.COUNT () = 0;

            FOR indx IN 1 .. many_rows.COUNT
            loop
               cntr := cntr + 1;
            end loop;
         END LOOP;

         DBMS_OUTPUT.put_line (cntr);

         CLOSE cur;
   END CASE;

   show_elapsed_time (approach);
END test_cursor_performance;
/

/* Try different approaches with optimization disabled. */

ALTER PROCEDURE test_cursor_performance
COMPILE plsql_optimize_level=0
/

BEGIN
   dbms_output.put_line ('No optimization...');
   
   test_cursor_performance ('implicit cursor for loop');

   test_cursor_performance ('explicit open, fetch, close');

   test_cursor_performance ('bulk fetch');
END;
/

/* Try different approaches with default optimization. */

ALTER PROCEDURE test_cursor_performance
COMPILE plsql_optimize_level=2
/

BEGIN
   DBMS_OUTPUT.put_line ('Default optimization...');

   test_cursor_performance ('implicit cursor for loop');

   test_cursor_performance ('explicit open, fetch, close');

   test_cursor_performance ('bulk fetch');
END;
/

/* Try different approaches with DEBUG enabled. */

ALTER PROCEDURE test_cursor_performance
COMPILE DEBUG
/

BEGIN
   DBMS_OUTPUT.put_line ('DEBUG enabled...');

   test_cursor_performance ('implicit cursor for loop');

   test_cursor_performance ('explicit open, fetch, close');

   test_cursor_performance ('bulk fetch');
END;
/

/*

On 11.2

No optimization...
"implicit cursor for loop" completed in: 2.95
"explicit open, fetch, close" completed in: 2.75
"bulk fetch" completed in: 1.33

Default optimization...
"implicit cursor for loop" completed in: 1.36
"explicit open, fetch, close" completed in: 2.75
"bulk fetch" completed in: 1.44

DEBUG enabled...
"implicit cursor for loop" completed in: 2.74
"explicit open, fetch, close" completed in: 2.78
"bulk fetch" completed in: 1.31

12.2

No optimization...
"implicit cursor for loop" completed in: 1.71
"explicit open, fetch, close" completed in: .99
"bulk fetch" completed in: .76

Default optimization...
"implicit cursor for loop" completed in: .5
"explicit open, fetch, close" completed in: .94
"bulk fetch" completed in: .49

DEBUG enabled...
"implicit cursor for loop" completed in: .94
"explicit open, fetch, close" completed in: .96
"bulk fetch" completed in: .5

*/