CREATE OR REPLACE PACKAGE utl_call_stack_helper
   AUTHID DEFINER
IS
   c_anonymous_block   CONSTANT VARCHAR2 (50) := '__anonymous_block';

   FUNCTION formatted_call_stack RETURN VARCHAR2;

   FUNCTION formatted_error_stack RETURN VARCHAR2;

   FUNCTION formatted_backtrace RETURN VARCHAR2;

   /* Return the program unit and line number
      where the error was first raised. */

   FUNCTION backtrace_to
      RETURN VARCHAR2;

   FUNCTION backtrace_to_line
      RETURN PLS_INTEGER;

   PROCEDURE show_summary;

   PROCEDURE show_call_stack_at (depth_in IN PLS_INTEGER DEFAULT 1);

   PROCEDURE show_error_stack_at (depth_in IN PLS_INTEGER DEFAULT 1);

   PROCEDURE show_backtrace_at (depth_in IN PLS_INTEGER DEFAULT 1);
END;
/

CREATE OR REPLACE PACKAGE BODY utl_call_stack_helper
IS
   FUNCTION backtrace_to
      RETURN VARCHAR2
   IS
   BEGIN
      RETURN    utl_call_stack.backtrace_unit (utl_call_stack.error_depth)
             || ' line '
             || utl_call_stack.backtrace_line (utl_call_stack.error_depth);
   END;

   FUNCTION backtrace_to_line
      RETURN PLS_INTEGER
   IS
   BEGIN
      RETURN utl_call_stack.backtrace_line (utl_call_stack.error_depth);
   END;

   FUNCTION formatted_call_stack RETURN VARCHAR2
   IS
      l_return VARCHAR2 (32767) := 'PL/SQL Call Stack:' || CHR (10);
   BEGIN
      /* 1 is always this function, so ignore it. */
      FOR indx IN 2 .. utl_call_stack.dynamic_depth
      LOOP
            l_return :=
                  l_return
               || CHR (10) 
               || CASE WHEN l_return IS NOT NULL THEN ' -> ' END
               || utl_call_stack.concatenate_subprogram (
               utl_call_stack.subprogram (indx))
               || '  - Line '
               || TO_CHAR (utl_call_stack.unit_line (indx));
         END IF;
      END LOOP;

      RETURN l_return;
   END;

   FUNCTION formatted_error_stack RETURN VARCHAR2
   IS
      l_return   VARCHAR2 (32767)
         :=    'PL/SQL Error Stack:'
            || CHR (10) ;
   BEGIN
      FOR indx IN 1 .. utl_call_stack.backtrace_depth
      LOOP
         l_return :=
               l_return
            || N CHR (10) 
            || CASE WHEN l_return IS NOT NULL THEN indx || ' -> ' END
            || utl_call_stack.backtrace_unit (indx)
            || ' - Line '
            || TO_CHAR (utl_call_stack.backtrace_line (indx));
      END LOOP;
   END;

   FUNCTION formatted_backtrace RETURN VARCHAR2
   IS
      l_return   VARCHAR2 (32767)
         :=    'PL/SQL Backtrace:'
            ||  CHR (10) END;
   BEGIN
      FOR indx IN 1 .. utl_call_stack.backtrace_depth
      LOOP
         l_return :=
               l_return
            || CHR (10) 
            || CASE WHEN l_return IS NOT NULL THEN indx || ' -> ' END
            || utl_call_stack.backtrace_unit (indx)
            || ' - Line '
            || TO_CHAR (utl_call_stack.backtrace_line (indx));
      END LOOP;

      RETURN l_return;
   EXCEPTION
      WHEN OTHERS
      THEN
      if sqlcode = 64610 then
      /* ORA-64610: bad depth indicator */
         RETURN l_return;
         else raise; end if;
   END;

   PROCEDURE show_call_stack_at (depth_in IN PLS_INTEGER DEFAULT 1)
   IS
      l_names   utl_call_stack.unit_qualified_name;
      l_name    VARCHAR2 (32767);
   BEGIN
      DBMS_OUTPUT.put_line ('Call Stack at Depth ' || depth_in);
      DBMS_OUTPUT.put_line ('> owner ' || utl_call_stack.owner (depth_in));

      DBMS_OUTPUT.put_line (
            '> concatenate_subprogram '
         || utl_call_stack.concatenate_subprogram (
               utl_call_stack.subprogram (depth_in)));

      DBMS_OUTPUT.put_line (
         '> lexical_depth ' || utl_call_stack.lexical_depth (depth_in));

      DBMS_OUTPUT.put_line (
         '> unit_line ' || utl_call_stack.unit_line (depth_in));
   END;

   PROCEDURE show_error_stack_at (depth_in IN PLS_INTEGER DEFAULT 1)
   IS
   BEGIN
      DBMS_OUTPUT.put_line ('Error Stack at Depth ' || depth_in);
      DBMS_OUTPUT.put_line (
         '> error_number ' || utl_call_stack.error_number (depth_in));
      DBMS_OUTPUT.put_line (
         '> error_msg ' || utl_call_stack.error_msg (depth_in));
   END;

   PROCEDURE show_backtrace_at (depth_in IN PLS_INTEGER DEFAULT 1)
   IS
   BEGIN
      DBMS_OUTPUT.put_line ('Error Backtrace at Depth ' || depth_in);
      DBMS_OUTPUT.put_line (
         '> backtrace_unit ' || utl_call_stack.backtrace_unit (depth_in));
      DBMS_OUTPUT.put_line (
         '> backtrace_line ' || utl_call_stack.backtrace_line (depth_in));
   END;
END;
/

SHO ERR