CREATE OR REPLACE TYPE strings_t IS TABLE OF VARCHAR2 (100);
/

CREATE OR REPLACE FUNCTION random_strings (count_in IN INTEGER)
   RETURN strings_t
   AUTHID DEFINER
IS
   l_strings   strings_t := strings_t ();
BEGIN
   l_strings.EXTEND (count_in);

   FOR indx IN 1 .. count_in
   LOOP
      l_strings (indx) := DBMS_RANDOM.string ('u', 10);
   END LOOP;

   RETURN l_strings;
END;
/

DECLARE
   l_strings   strings_t := random_strings (5);
BEGIN
   FOR indx IN 1 .. l_strings.COUNT
   LOOP
      DBMS_OUTPUT.put_line (l_strings (indx));
   END LOOP;
END;
/

SELECT COLUMN_VALUE my_string FROM TABLE (random_strings (5))
/

SELECT e.last_name
  FROM TABLE (random_strings (3)) rs, employees e
 WHERE LENGTH (e.last_name) <= LENGTH (COLUMN_VALUE)
/

SELECT e.last_name
  FROM employees e
 WHERE ROWNUM < 4
UNION ALL
SELECT rs.COLUMN_VALUE
  FROM TABLE (random_strings (3)) rs
/

BEGIN
   FOR rec IN (SELECT COLUMN_VALUE my_string FROM TABLE (random_strings (5)))
   LOOP
      DBMS_OUTPUT.put_line (rec.my_string);
   END LOOP;
END;
/

/* Cannot use with local/nested function */

DECLARE
   FUNCTION my_random_strings (count_in IN INTEGER)
      RETURN strings_t
   IS
      l_strings   strings_t := strings_t ();
   BEGIN
      l_strings.EXTEND (count_in);

      FOR indx IN 1 .. count_in
      LOOP
         l_strings (indx) := DBMS_RANDOM.string ('u', 10);
      END LOOP;

      RETURN l_strings;
   END;
BEGIN
   FOR rec
      IN (SELECT COLUMN_VALUE my_string FROM TABLE (my_random_strings (5)))
   LOOP
      DBMS_OUTPUT.put_line (rec.my_string);
   END LOOP;
END;
/

/* Cannot use a collection type defined in package spec (if not pipelined) */

CREATE OR REPLACE PACKAGE tf
   AUTHID DEFINER
IS
   TYPE strings_t IS TABLE OF VARCHAR2 (100);

   FUNCTION random_strings (count_in IN INTEGER)
      RETURN strings_t;

   FUNCTION just_number
      RETURN NUMBER;
END;
/

CREATE OR REPLACE PACKAGE BODY tf
IS
   FUNCTION random_strings (count_in IN INTEGER)
      RETURN strings_t
   IS
      l_strings   strings_t := strings_t ();
   BEGIN
      l_strings.EXTEND (count_in);

      FOR indx IN 1 .. count_in
      LOOP
         l_strings (indx) := DBMS_RANDOM.string ('u', 10);
      END LOOP;

      RETURN l_strings;
   END;

   FUNCTION just_number
      RETURN NUMBER
   IS
   BEGIN
      RETURN 1;
   END;
END;
/

SELECT tf.just_number FROM DUAL
/

SELECT COLUMN_VALUE my_string FROM TABLE (tf.random_strings (5))
/

/* But can use a packaged function the returns a schema-level type */

CREATE OR REPLACE TYPE strings_t IS TABLE OF VARCHAR2 (100);
/

CREATE OR REPLACE PACKAGE tf
   AUTHID DEFINER
IS
   FUNCTION random_strings (count_in IN INTEGER)
      RETURN strings_t;
END;
/

CREATE OR REPLACE PACKAGE BODY tf
IS
   FUNCTION random_strings (count_in IN INTEGER)
      RETURN strings_t
   IS
      l_strings   strings_t := strings_t ();
   BEGIN
      l_strings.EXTEND (count_in);

      FOR indx IN 1 .. count_in
      LOOP
         l_strings (indx) := DBMS_RANDOM.string ('u', 10);
      END LOOP;

      RETURN l_strings;
   END;
END;
/

SELECT COLUMN_VALUE my_string FROM TABLE (tf.random_strings (5))
/

/* How often is function called? */

CREATE OR REPLACE FUNCTION random_strings (count_in IN INTEGER)
   RETURN strings_t
   AUTHID DEFINER
IS
   l_strings   strings_t := strings_t ();
BEGIN
   DBMS_OUTPUT.PUT_LINE ('random_strings called');
   
   l_strings.EXTEND (count_in);

   FOR indx IN 1 .. count_in
   LOOP
      l_strings (indx) := DBMS_RANDOM.string ('u', 10);
   END LOOP;

   RETURN l_strings;
END;
/

BEGIN
   FOR rec IN (SELECT e.last_name
                 FROM employees e, TABLE (random_strings (5)) rs
                WHERE ROWNUM < 3)
   LOOP
      DBMS_OUTPUT.put_line (rec.last_name);
   END LOOP;
END;
/

/* Left correlation behavior */

DROP TABLE things
/

CREATE TABLE things
(
   thing_id     NUMBER,
   thing_name   VARCHAR2 (100)
)
/

BEGIN
   INSERT INTO things VALUES (1, 'Thing 1');
   INSERT INTO things VALUES (2, 'Thing 2');
   COMMIT;
END;
/

CREATE OR REPLACE TYPE numbers_t IS TABLE OF NUMBER
/

CREATE OR REPLACE FUNCTION more_numbers (id_in IN NUMBER)
   RETURN numbers_t
   AUTHID DEFINER
IS
   l_numbers   numbers_t := numbers_t ();
BEGIN
   l_numbers.EXTEND (id_in * 5);

   FOR indx IN 1 .. id_in * 5
   LOOP
      l_numbers (indx) := indx;
   END LOOP;

   DBMS_OUTPUT.put_line ('more numbers');
   RETURN l_numbers;
END;
/

BEGIN
   FOR rec IN (SELECT th.thing_name, t.COLUMN_VALUE thing_number
                 FROM things th, TABLE (more_numbers (th.thing_id)) t)
   LOOP
      DBMS_OUTPUT.put_line ('more numbers ' || rec.thing_number);
   END LOOP;
END;
/

/* Function can be schema level or in a package spec */

CREATE OR REPLACE TYPE strings_t IS TABLE OF VARCHAR2 (100);
/

CREATE OR REPLACE PACKAGE tf AUTHID DEFINER
IS
   FUNCTION strings RETURN strings_t;
END;
/

CREATE OR REPLACE PACKAGE BODY tf
IS
   FUNCTION strings RETURN strings_t
   IS
   BEGIN
      RETURN strings_t ('abc');
   END;
END;
/

SELECT COLUMN_VALUE my_string FROM TABLE (tf.strings)
/
e
DECLARE
   FUNCTION nested_strings (count_in IN INTEGER) RETURN strings_t
   IS
   BEGIN
      RETURN strings_t ('abc');
   END;
BEGIN
   FOR rec IN (SELECT * FROM TABLE (nested_strings()))
   LOOP
      DBMS_OUTPUT.PUT_LINE (rec.COLUMN_VALUE);
   END LOOP;
END;
/