DROP TYPE ticker_ot FORCE;
DROP TYPE tickers_nt FORCE;
DROP TYPE stock_ot FORCE;
DROP TYPE stocks_nt FORCE;
DROP TABLE stocks;
DROP TABLE tickers;

CREATE OR REPLACE PROCEDURE show_pga_memory (context_in        IN VARCHAR2,
                                             pga_only          IN BOOLEAN DEFAULT FALSE,
                                             compare_to_last   IN BOOLEAN DEFAULT TRUE)
   AUTHID DEFINER                                             
IS
   l_value   NUMBER;
BEGIN
   SELECT s.VALUE
     INTO l_value
     FROM sys.v_$sesstat s, sys.v_$statname n, sys.v_$session sh
    WHERE     s.statistic# = n.statistic#
          AND s.sid = sh.sid
          AND sh.audsid = USERENV ('SESSIONID')
          AND n.name = 'session pga memory';

   DBMS_OUTPUT.put_line ('PGA level ' || context_in || ': ' || l_value);
   DBMS_SESSION.free_unused_user_memory;
END;
/

CREATE TABLE stocks
(
   ticker        VARCHAR2 (20),
   trade_date    DATE,
   open_price    NUMBER,
   close_price   NUMBER
)
/

BEGIN
   FOR indx IN 1 .. 100000
   LOOP
      -- Might as well be optimistic!
      INSERT INTO stocks
           VALUES ('STK' || indx,
                   SYSDATE,
                   indx,
                   indx + 15);
   END LOOP;

   COMMIT;
END;
/

CREATE TABLE tickers
(
   ticker      VARCHAR2 (20),
   pricedate   DATE,
   pricetype   VARCHAR2 (1),
   price       NUMBER
)
/

CREATE TYPE ticker_ot AS OBJECT
(
   ticker VARCHAR2 (20),
   pricedate DATE,
   pricetype VARCHAR2 (1),
   price NUMBER
);
/

CREATE TYPE tickers_nt AS TABLE OF ticker_ot;
/

CREATE OR REPLACE PACKAGE stock_mgr
   AUTHID DEFINER
IS
   TYPE stocks_rc IS REF CURSOR
      RETURN stocks%ROWTYPE;
END stock_mgr;
/

CREATE OR REPLACE FUNCTION stockpivot_bulk (rows_in stock_mgr.stocks_rc)
   RETURN tickers_nt
   AUTHID DEFINER
IS
   /* Container for rows fetched from the cursor variable. */
   TYPE stocks_aat IS TABLE OF stocks%ROWTYPE
      INDEX BY PLS_INTEGER;

   l_stocks    stocks_aat;

   /* The nested table that will be returned. */
   l_doubled   tickers_nt := tickers_nt ();
BEGIN
   LOOP
      /* Fetch next 100 rows. */
      FETCH rows_in BULK COLLECT INTO l_stocks LIMIT 100;

      EXIT WHEN l_stocks.COUNT = 0;

      FOR l_row IN 1 .. l_stocks.COUNT
      LOOP
         l_doubled.EXTEND;
         l_doubled (l_doubled.LAST) :=
            ticker_ot (l_stocks (l_row).ticker,
                       l_stocks (l_row).trade_date,
                       'O',
                       l_stocks (l_row).open_price);

         l_doubled.EXTEND;
         l_doubled (l_doubled.LAST) :=
            ticker_ot (l_stocks (l_row).ticker,
                       l_stocks (l_row).trade_date,
                       'C',
                       l_stocks (l_row).close_price);
      END LOOP;
   END LOOP;

   CLOSE rows_in;
   
   show_pga_memory ('Inside Bulk TF');

   RETURN l_doubled;
END;
/

CREATE OR REPLACE FUNCTION stockpivot_nobulk (rows_in stock_mgr.stocks_rc)
   RETURN tickers_nt
   AUTHID DEFINER
IS
   l_stock     stocks%ROWTYPE;
   l_doubled   tickers_nt := tickers_nt ();
BEGIN
   LOOP
      FETCH rows_in INTO l_stock;

      EXIT WHEN rows_in%NOTFOUND;

      l_doubled.EXTEND;
      l_doubled (l_doubled.LAST) :=
         ticker_ot (l_stock.ticker,
                    l_stock.trade_date,
                    'O',
                    l_stock.open_price);

      l_doubled.EXTEND;
      l_doubled (l_doubled.LAST) :=
         ticker_ot (l_stock.ticker,
                    l_stock.trade_date,
                    'C',
                    l_stock.close_price);
   END LOOP;

   CLOSE rows_in;
   
   show_pga_memory ('Inside Non-Bulk TF');

   RETURN l_doubled;
END;
/

DECLARE
   l_count INTEGER;
BEGIN
   show_pga_memory ('Before Bulk TF');

   SELECT COUNT (*) INTO l_count FROM TABLE (stockpivot_bulk (CURSOR (SELECT * FROM stocks)));

   DBMS_OUTPUT.put_line ('TF count: ' || l_count);
   show_pga_memory ('After  Bulk TF');
END;
/

DECLARE
   l_count INTEGER;
BEGIN
   show_pga_memory ('Before Non-bulk TF');

   SELECT COUNT (*) INTO l_count FROM TABLE (stockpivot_nobulk (CURSOR (SELECT * FROM stocks)));

   DBMS_OUTPUT.put_line ('TF count: ' || l_count);
   show_pga_memory ('After  Non-bulk TF');
END;
/

CREATE OR REPLACE FUNCTION stockpivot_nopl (rows_in stock_mgr.stocks_rc)
   RETURN tickers_nt
   AUTHID DEFINER
IS
   /* Container for rows fetched from the cursor variable. */
   TYPE stocks_aat IS TABLE OF stocks%ROWTYPE
      INDEX BY PLS_INTEGER;

   l_stocks    stocks_aat;

   /* The nested table that will be returned. */
   l_doubled   tickers_nt := tickers_nt ();
BEGIN
   LOOP
      /* Fetch next 100 rows. */
      FETCH rows_in BULK COLLECT INTO l_stocks LIMIT 100;

      EXIT WHEN l_stocks.COUNT = 0;

      FOR l_row IN 1 .. l_stocks.COUNT
      LOOP
         l_doubled.EXTEND;
         l_doubled (l_doubled.LAST) :=
            ticker_ot (l_stocks (l_row).ticker,
                       l_stocks (l_row).trade_date,
                       'O',
                       l_stocks (l_row).open_price);

         l_doubled.EXTEND;
         l_doubled (l_doubled.LAST) :=
            ticker_ot (l_stocks (l_row).ticker,
                       l_stocks (l_row).trade_date,
                       'C',
                       l_stocks (l_row).close_price);
      END LOOP;
   END LOOP;

   CLOSE rows_in;

   RETURN l_doubled;
END;
/

CREATE OR REPLACE FUNCTION stockpivot_pl (rows_in stock_mgr.stocks_rc)
   RETURN tickers_nt
   PIPELINED
   AUTHID DEFINER
IS
   TYPE stocks_aat IS TABLE OF stocks%ROWTYPE
      INDEX BY PLS_INTEGER;

   l_stocks   stocks_aat;
BEGIN
   LOOP
      /* Fetch next 100 rows. */
      FETCH rows_in BULK COLLECT INTO l_stocks LIMIT 100;

      EXIT WHEN l_stocks.COUNT = 0;

      FOR l_row IN 1 .. l_stocks.COUNT
      LOOP
         PIPE ROW (ticker_ot (l_stocks (l_row).ticker,
                              l_stocks (l_row).trade_date,
                              'O',
                              l_stocks (l_row).open_price));

         PIPE ROW (ticker_ot (l_stocks (l_row).ticker,
                              l_stocks (l_row).trade_date,
                              'C',
                              l_stocks (l_row).close_price));
      END LOOP;
   END LOOP;

   CLOSE rows_in;

   RETURN;
END;
/

/* Compare PGA memory for pipelined and non-pipelined function */

BEGIN
   DELETE FROM tickers;

   COMMIT;

   show_pga_memory ('Before Pipelined TF');

   INSERT INTO tickers
      SELECT * FROM TABLE (stockpivot_pl (CURSOR (SELECT * FROM stocks)));

   show_pga_memory ('After  Pipelined TF');
END;
/

BEGIN
   DELETE FROM tickers;

   COMMIT;

   show_pga_memory ('Before Non-pipelined');

   INSERT INTO tickers
      SELECT * FROM TABLE (stockpivot_nopl (CURSOR (SELECT * FROM stocks)));

   show_pga_memory ('After  Non-pipelined');
END;
/