-- sparse and dense

declare
   type colors_t is table of varchar2(100) index by pls_integer;
   l_colors_dense    colors_t;
   l_colors_sparse   colors_t;
begin
   l_colors_dense(100) := 'Blue';
   l_colors_dense(101) := 'Green';
   
   l_colors_sparse(1) := 'Red';
   l_colors_sparse(3) := 'Yellow';
end;   
/

-- basic for loop iteration through dense array

declare
   type colors_t is table of varchar2(100) index by pls_integer;
   l_colors colors_t := colors_t (1 => 'Blue', 2 => 'Green', 3 => 'Yellow');
begin
   for indx in 1 .. l_colors.count
   loop
      dbms_output.put_line (l_colors (indx));
   end loop;  
end;   
/

-- for loop with sparse, a no-no

declare
   type colors_t is table of varchar2(100) index by pls_integer;
   l_colors colors_t;
begin
   l_colors (100) := 'Blue';
   l_colors (101) := 'Green';

   for indx in 1 .. l_colors.count
   loop
      dbms_output.put_line (l_colors  (indx));
   end loop;  
end;   
/

-- Easier to see why here

declare
   type colors_t is table of varchar2(100) index by pls_integer;
   l_colors colors_t;
begin
   l_colors (100) := 'Blue';
   l_colors (101) := 'Green';

   dbms_output.put_line (l_colors (1));
   dbms_output.put_line (l_colors (2));
end;   
/

-- first and last instead of 1 and count

declare
   type colors_t is table of varchar2(100) index by pls_integer;
   l_colors colors_t;
begin
   l_colors(100) := 'Blue';
   l_colors(101) := 'Green';

   for indx in l_colors.first .. l_colors.last
   loop
      dbms_output.put_line (l_colors (indx));
   end loop;  
end;   
/


declare
   type colors_t is table of varchar2(100) index by pls_integer;
   l_colors colors_t;
begin
   l_colors(100) := 'Blue';
   l_colors(101) := 'Green';

   if l_colors.count > 0
   then
   for indx in l_colors.first .. l_colors.last
   loop
      dbms_output.put_line (l_colors (indx));
   end loop;  
   end if;
end;   

-- protection from value_error exception

declare
   type colors_t is table of varchar2(100) index by pls_integer;
   l_colors colors_t;
begin
   for indx in nvl (l_colors.first, 1) .. nvl (l_colors.last, 0)
   loop
      dbms_output.put_line (l_colors (indx));
   end loop;  
end; 
/

-- avoiding no_data_found ... the wrong way

declare
   type colors_t is table of varchar2(100) index by pls_integer;
   l_colors colors_t;
begin
   l_colors(100) := 'Blue';
   l_colors(1000) := 'Green';
   l_colors(1000000000) := 'Green';

   for indx in l_colors.first .. l_colors.last
   loop
      begin
         dbms_output.put_line (l_colors (indx));
      exception when no_data_found then null;
      end;
   end loop;  
end;   
/

declare
   type colors_t is table of varchar2(100) index by pls_integer;
   l_colors colors_t;
begin
   l_colors(100) := 'Blue';
   l_colors(1000) := 'Green';

   for indx in l_colors.first .. l_colors.last
   loop
      if l_colors.exists (indx) 
      then
         dbms_output.put_line (l_colors (indx));
      end if;
   end loop;  
end;   
/

-- not exactly concise but correct looping through sparse collection

declare
   type colors_t is table of varchar2(100) index by pls_integer;
   l_colors colors_t;
   l_index pls_integer;
begin
   l_colors(100) := 'Blue';
   l_colors(1000) := 'Green';

   l_index := l_colors.first;
   while l_index is not null
   loop
      dbms_output.put_line (l_colors (l_index));
      l_index := l_colors.next (l_index);
   end loop;  
end;   
/

-- and then there is SQL

create or replace type list_of_names_t is table of varchar2(100);
/

declare
   happyfamily list_of_names_t := list_of_names_t('Kim', 'Kourtney', 'Khloe', 'Rob');
begin
   for rec in (
      select column_value family_name
        from table ( happyfamily )
        union 
        select last_name from hr.employees
       order by family_name
   ) loop
      DBMS_OUTPUT.put_line(rec.family_name);
   end loop;
end;
/

-- and in 21.1 and higher

declare
   type colors_t is table of varchar2(100) index by pls_integer;
   l_colors colors_t;
begin
   l_colors(100) := 'Blue';
   l_colors(1000) := 'Green';

   for indx in indices of l_colors
   loop
      dbms_output.put_line (l_colors (indx));
   end loop;  
end;   
/


