CREATE OR REPLACE PROCEDURE create_custom_workout (
   user_id_in            IN     INTEGER,
   topic_id_in           IN     INTEGER DEFAULT NULL,
   max_minutes_in        IN     INTEGER DEFAULT NULL,
   difficulty_range_in   IN     INTEGER DEFAULT NULL,
   months_unplayed_in    IN     INTEGER DEFAULT 3,
   question_ids_in       IN     VARCHAR2 DEFAULT NULL,
   question_query_in     IN     VARCHAR2 DEFAULT NULL,
   title_in              IN     VARCHAR2 DEFAULT NULL,
   workout_id_out           OUT INTEGER)
IS
   c_target_time         CONSTANT NUMBER := max_minutes_in;
   c_difficulty_string   CONSTANT qdb_config.maxvarchar2
      := dg_workout_mgr.difficulty_range_where_clause (difficulty_range_in) ;
   l_title                        qdb_config.maxvarchar2;
   l_counter                      INTEGER := 1;
   l_total_time                   NUMBER := 0;
   l_ce_id                        INTEGER;
   l_quizzes_available            BOOLEAN DEFAULT FALSE;

   PROCEDURE get_text_for_title
   IS
   BEGIN
      CASE
         WHEN title_in IS NOT NULL
         THEN
            l_title := title_in;
         ELSE
            SELECT t.text || ' in ' || max_minutes_in || ' Minutes'
              INTO l_title
              FROM qdb_difficulty_levels dr, qdb_topics t
             WHERE     dr.difficulty_id = difficulty_range_in
                   AND t.topic_id = topic_id_in;
      END CASE;
   END;

   PROCEDURE load_up_activities
   IS
      resources_rec   ov.ov_resources_external%ROWTYPE;
      resources_cv    SYS_REFCURSOR;

      l_activities    dg_workout_activities_nt := dg_workout_activities_nt ();

      l_query         qdb_config.maxvarchar2;

      PROCEDURE add_one_resource (
         resource_in     IN     ov.ov_resources_external%ROWTYPE,
         total_time_io   IN OUT NUMBER,
         counter_io      IN OUT NUMBER)
      IS
      BEGIN
         l_activities.EXTEND;

         l_activities (l_activities.LAST) :=
            dg_workout_mgr.activity_from_resource (
               resource_in            => resource_in,
               activity_sequence_in   => counter_io);

         total_time_io :=
              total_time_io
            + dg_workout_mgr.minutes_for_activity_type (
                 duration_mins_in      => resource_in.duration_mins,
                 activity_type_id_in   => dg_workout_mgr.act_type_from_res_usage (
                                            resource_in.resource_usage));

         counter_io := counter_io + 1;
      END;

      FUNCTION resources_for_usage_query (usage_in              IN VARCHAR2,
                                          difficulty_range_in   IN VARCHAR2)
         RETURN VARCHAR2
      IS
         l_query   qdb_config.maxvarchar2;
      BEGIN
         l_query :=
            CASE
               WHEN question_query_in IS NOT NULL
               THEN
                  
                   'SELECT r.*
             FROM ov.ov_resources_external r, qdb_questions q
            WHERE r.external_resource_key = q.question_id
              and r.resource_usage = ''Quiz''
              and q.question_id IN ('
                  || question_query_in
                  || ') order by r.difficulty_number'
                  
               WHEN question_ids_in IS NOT NULL
               THEN
                     'SELECT r.*
             FROM ov.ov_resources_external r, qdb_questions q
            WHERE r.external_resource_key = q.question_id
              and r.resource_usage = ''Quiz''
              and q.question_id IN ('
                  || question_ids_in
                  || ') order by r.difficulty_number'
               ELSE
                     '
           SELECT r.*
             FROM ov.ov_resources_external r
            WHERE r.topic_id IN (SELECT tt.topic_id
                                        FROM ov.ov_topics tt
                                     CONNECT BY PRIOR tt.topic_id = tt.parent_topic_id
                                       START WITH tt.topic_id =  :topic_id) AND r.'
                  || difficulty_range_in
                  || ' AND duration_mins < 15 AND r.minimum_version_id IS NOT NULL AND r.resource_usage IN ('
                  || usage_in
                  || ') AND '
                  || CASE
                        WHEN months_unplayed_in IS NULL
                        THEN
                           ':user_id IS NOT NULL AND :months IS NULL'
                        ELSE
                           'external_resource_key not in (
                         SELECT qz.question_id
                           FROM qdb_compev_answers_v eva, qdb_quizzes qz
                          WHERE     eva.user_id = :user_id
                                AND eva.comp_event_id = qz.comp_event_id
                                AND eva.ended_on >= ADD_MONTHS (SYSDATE, -1 * :months))'
                     END
                  || ' ORDER BY CASE WHEN r.topic_id = :topic_id THEN 1 ELSE 0 END, difficulty_number asc'
            END;

         IF qdb_utilities.trace_is_on
         THEN
            qdb_utilities.trace_activity ('create_custom_workout query',
                                          l_query);
         END IF;

         DBMS_OUTPUT.put_line (l_query);
         RETURN l_query;
      END;
   BEGIN
      l_query :=
         resources_for_usage_query (
            usage_in              => 'Quiz', -- usage_string_for_content_type (c_only_quizzes),
            difficulty_range_in   => c_difficulty_string);

      IF question_ids_in IS NOT NULL OR question_query_in IS NOT NULL
      THEN
         OPEN resources_cv FOR l_query;
      ELSE
         OPEN resources_cv FOR l_query
            USING topic_id_in,
                  user_id_in,
                  months_unplayed_in,
                  topic_id_in;
      END IF;

      LOOP
         FETCH resources_cv INTO resources_rec;

         IF qdb_utilities.trace_is_on
         THEN
            qdb_utilities.trace_activity ('create_custom_workout add question',
                                          resources_rec.external_resource_key);
         END IF;

         EXIT WHEN    l_total_time >= c_target_time
                   OR l_total_time + resources_rec.duration_mins >
                         c_target_time
                   OR resources_cv%NOTFOUND;

         add_one_resource (resource_in     => resources_rec,
                           total_time_io   => l_total_time,
                           counter_io      => l_counter);
      END LOOP;

      CLOSE resources_cv;

      FORALL indx IN 1 .. l_activities.COUNT
         INSERT INTO dg_workout_activities (workout_id,
                                            activity_type_id,
                                            resource_id,
                                            activity_sequence)
              VALUES (workout_id_out,
                      l_activities (indx).activity_type_id,
                      l_activities (indx).resource_id,
                      l_activities (indx).activity_sequence);

      l_quizzes_available := l_activities.COUNT > 0;
   END;
BEGIN
   get_text_for_title;

   /* Ccreate the workout, then add the user */
   INSERT INTO dg_workouts (workout_name,
                            workout_type,
                            start_date,
                            end_date,
                            difficulty_id,
                            difficulty_range,
                            description,
                            topic_id,
                            approved_by_user_id,
                            author_id,
                            comp_event_id,
                            max_minutes)
        VALUES (l_title,
                'SHARED',
                SYSDATE,
                NULL,
                difficulty_range_in,
                difficulty_range_in,
                'A workout created just for you: ' || l_title,
                topic_id_in,
                qdb_config.admin_user_id,
                user_id_in,
                NULL,
                max_minutes_in)
     RETURNING workout_id
          INTO workout_id_out;

   load_up_activities;

   IF l_quizzes_available
   THEN
      dg_workout_mgr.create_workout_ce (workout_id_in       => workout_id_out,
                                        comp_event_id_out   => l_ce_id,
                                        start_date_in       => SYSDATE,
                                        end_date_in         => NULL);

      dg_workout_mgr.add_user_to_workout (p_workout_id   => workout_id_out,
                                          p_user_id      => user_id_in);
   ELSE
      raise_application_error (
         -20000,
         'Unable to find quizzes for your criteria. Please try again.');
   END IF;
END;
/

DECLARE
   id   INTEGER;
BEGIN
   create_custom_workout (
      user_id_in => 5,
      difficulty_range_in => 123,
      title_in => 'Santa Claus SQL',
      question_query_in   => q'[select question_id q from qdb_questions q, 
qdb_topics t where upper(question) like '%SANTA%' and q.topic_id = t.topic_id and t.domain_id = 21]',
      workout_id_out      => id);
      exception when others
      then 
      dbms_output.put_line (dbms_utility.format_error_stack);
      dbms_output.put_line (dbms_utility.format_error_backtrace);
END;
/