/* Schema-level procedure: this means that anyone with */

CREATE OR REPLACE PROCEDURE process_one_row (table_in   IN VARCHAR2,
                                             where_in   IN VARCHAR2)
IS
   l_block   VARCHAR2 (32767)
      :=    'DECLARE l_row '
         || table_in
         || '%ROWTYPE; BEGIN SELECT * INTO l_row FROM '
         || table_in
         || ' WHERE '
         || where_in
         || '; /* internal_pkg.process_data (l_row); */'
         || 'END;';
BEGIN
   EXECUTE IMMEDIATE l_block;
EXCEPTION
   WHEN OTHERS
   THEN
      Raise_Application_Error (
         -20000,
            'Unable to execute dynamic statement for table-where '
         || table_in
         || '-'
         || where_in
         || ' = '
         || l_block
         || ' Error: '
         || SQLERRM);
END;
/

BEGIN
   process_one_row ('EMPLOYEES', 'employee_id=101');
END;
/

BEGIN
   process_one_row (
      'EMPLOYEES',
      'employee_id=101; 
      EXECUTE IMMEDIATE ''CREATE TABLE nasty_data (mycol NUMBER)''');
END;
/

BEGIN
   process_one_row ('EMPLOYEES',
                    'employee_id=101; 
      EXECUTE IMMEDIATE 
         ''CREATE PROCEDURE backdoor (str VARCHAR2) 
          AS BEGIN EXECUTE IMMEDIATE str; END;''');
END;
/

/* With dynamic SQL, you might be threatened by strings like: 

EXEC reports.show_col('lname','customers');
EXEC reports.show_col('ssn','customers');

-- lets get some sensitive information
EXEC reports.show_col('lname','customers where 5<1 union select distinct name from user_source --');
EXEC reports.show_col('lname','customers where 5<1 union select username from all_users --');

*/

CREATE OR REPLACE PACKAGE dynamic_hr
AS
   PROCEDURE process_one_row (table_in IN VARCHAR2, where_in IN VARCHAR2);
END;
/

CREATE OR REPLACE PACKAGE BODY dynamic_hr
AS
   PROCEDURE process_one_row (table_in IN VARCHAR2, where_in IN VARCHAR2)
   IS
      l_block   VARCHAR2 (32767)
         :=    'DECLARE l_row '
            || table_in
            || '%ROWTYPE; BEGIN SELECT * INTO l_row FROM '
            || table_in
            || ' WHERE '
            || where_in
            || '; internal_process_data (l_row); '
            || 'END;';
   BEGIN
      EXECUTE IMMEDIATE l_block;
   EXCEPTION
      WHEN OTHERS
      THEN
         Raise_Application_Error (
            -20000,
               'Unable to execute dynamic statement for table-where '
            || table_in
            || '-'
            || where_in
            || ' = '
            || l_block
            || ' Error: '
            || SQLERRM);
   END;
END;
/

/*
   One way to minimize the danger of injection is to avoid the
   completely generic WHERE clause and instead code for
   specific, anticipated variations.
*/

CREATE OR REPLACE PACKAGE dynamic_hr
AS
   PROCEDURE process_one_row (table_in    IN VARCHAR2,
                              value1_in   IN VARCHAR2,
                              value2_in   IN DATE);
END;
/

CREATE OR REPLACE PACKAGE BODY dynamic_hr
AS
   PROCEDURE process_one_row (table_in    IN VARCHAR2,
                              value1_in   IN VARCHAR2,
                              value2_in   IN DATE)
   IS
      l_where   VARCHAR2 (32767);
   BEGIN
      IF table_in = 'EMPLOYEES'
      THEN
         l_where := 'last_name = :name AND hire_date < :hdate';
      ELSIF table_in = 'DEPARTMENTS'
      THEN
         l_where := 'name LIKE :name AND incorporation_date = :hdate';
      ELSE
         raise_application_error (
            -20000,
            'Invalid table name for process_one_row: ' || table_in);
      END IF;

      EXECUTE IMMEDIATE
            'DECLARE l_row '
         || table_in
         || '%ROWTYPE; BEGIN SELECT * INTO l_row FROM '
         || table_in
         || ' WHERE '
         || l_where
         || '; END;'
         USING value1_in, value2_in;
   EXCEPTION
      WHEN OTHERS
      THEN
         Raise_Application_Error (
            -20000,
               'Unable to execute dynamic statement for table-where '
            || table_in
            || '-'
            || where_in
            || ' = '
            || l_block
            || ' Error: '
            || SQLERRM);
   END;
END;
/

/* Use dbms_assert to validate DB object and schema names */

CREATE OR REPLACE PACKAGE BODY dynamic_hr
AS
   PROCEDURE process_one_row (table_in    IN VARCHAR2,
                              value1_in   IN VARCHAR2,
                              value2_in   IN DATE)
   IS
      l_where   VARCHAR2 (32767);
      l_table   VARCHAR2 (32767);
   BEGIN
      l_table := SYS.DBMS_ASSERT.sql_object_name (table_in);

      IF l_table = 'EMPLOYEES'
      THEN
         l_where := 'last_name = :name AND hire_date < :hdate';
      ELSIF l_table = 'DEPARTMENTS'
      THEN
         l_where := 'name LIKE :name AND incorporation_date = :hdate';
      ELSE
         raise_application_error (
            -20000,
            'Invalid table name for process_one_row: ' || table_in);
      END IF;

      EXECUTE IMMEDIATE
            'DECLARE l_row '
         || l_table
         || '%ROWTYPE; BEGIN SELECT * INTO l_row FROM '
         || l_table
         || ' WHERE '
         || l_where
         || ';'
         || 'END;'
         USING value1_in, value2_in;
   EXCEPTION
      WHEN OTHERS
      THEN
         Raise_Application_Error (
            -20000,
               'Unable to execute dynamic statement for table-where '
            || table_in
            || '-'
            || where_in
            || ' = '
            || l_block
            || ' Error: '
            || SQLERRM);
   END;
END;
/

/* Invoker rights for stored program unit dynamic SQL */

CREATE OR REPLACE PACKAGE dynamic_hr
   AUTHID CURRENT_USER
AS
   PROCEDURE process_one_row (table_in    IN VARCHAR2,
                              value1_in   IN VARCHAR2,
                              value2_in   IN DATE);
END;
/

/* Avoid passed back sensitive information through error reports. */

CREATE OR REPLACE PACKAGE BODY dynamic_hr
AS
   PROCEDURE process_one_row (table_in    IN VARCHAR2,
                              value1_in   IN VARCHAR2,
                              value2_in   IN DATE)
   IS
      l_where   VARCHAR2 (32767);
      l_table   VARCHAR2 (32767);
   BEGIN
      l_table := SYS.DBMS_ASSERT.sql_object_name (table_in);

      IF l_table = 'EMPLOYEES'
      THEN
         l_where := 'last_name = :name AND hire_date < :hdate';
      ELSIF l_table = 'DEPARTMENTS'
      THEN
         l_where := 'name LIKE :name AND incorporation_date = :hdate';
      ELSE
         raise_application_error (
            -20000,
            'Invalid table name for process_one_row: ' || table_in);
      END IF;

      EXECUTE IMMEDIATE
            'DECLARE l_row '
         || l_table
         || '%ROWTYPE; BEGIN SELECT * INTO l_row FROM '
         || l_table
         || ' WHERE '
         || l_where
         || ';'
         || 'END;'
         USING value1_in, value2_in;
   EXCEPTION
      WHEN OTHERS
      THEN
         DECLARE
            l_id   INTEGER;
         BEGIN
            log_error (help_desk_id_out => l_id);

            Raise_Application_Error (
               -20000,
               'Error processing request. Contact help desk with ID ' || l_id);
         END;
   END;
END;
/

/* ACCESSIBLE_BY */

CREATE OR REPLACE PACKAGE dynamic_hr
   AUTHID CURRENT_USER
   ACCESSIBLE BY (public_facing_pkg)
AS
   PROCEDURE process_one_row (table_in    IN VARCHAR2,
                              value1_in   IN VARCHAR2,
                              value2_in   IN DATE);
END;
/
   