create or replace package tmr is 
   procedure start_timing (p_title in varchar2); 
   function elapsed return number; 
   procedure show_elapsed; 
end tmr; 
/

create or replace package body tmr is 
   last_timing number := null; 
 
   procedure start_timing (p_title in varchar2) is 
   begin 
      last_timing := DBMS_UTILITY.GET_TIME; 
      dbms_output.put_line (p_title);
   end; 
 
   function elapsed return number is 
   begin 
      return ( mod(DBMS_UTILITY.GET_TIME - last_timing + POWER(2, 32), POWER(2, 32)) ); 
   end; 
 
   procedure show_elapsed is 
   begin 
      DBMS_OUTPUT.PUT_LINE(elapsed); 
   end; 
end tmr; 
/

create table my_users (
   id             number
      generated always as identity,
   nm             varchar2(100),
   email_address  varchar2(1000) unique
)
/

begin
   insert into my_users (
      nm,
      email_address
   )
      select 'Steven ' || level,
             'steven'
             || level
             || '@insum.ca'
        from dual connect by
         level < 10000;
   commit;
end;
/

-- Try to query then insert if failure

create or replace function insert_and_or_return_user1 (
   p_nm             in  my_users.nm%type,
   p_email_address  in  my_users.email_address%type
) return my_users.id%type
   authid definer
is
   l_id my_users.id%type;
begin
   select id
     into l_id
     from my_users
    where email_address = p_email_address;

   return l_id;
exception
   when no_data_found then
      insert into my_users (
         nm,
         email_address
      ) values (
         p_nm,
         p_email_address
      ) return id into l_id;
      return l_id;
end;
/

-- Try to insert then query if failure

create or replace function insert_and_or_return_user2 (
   p_nm             in  my_users.nm%type,
   p_email_address  in  my_users.email_address%type
) return my_users.id%type
   authid definer
is
   l_id my_users.id%type;
begin
   insert into my_users (
      nm,
      email_address
   ) values (
      p_nm,
      p_email_address
   ) return id into l_id;
   return l_id;
exception
   when dup_val_on_index then
      select id
        into l_id
        from my_users
       where email_address = p_email_address;

      return l_id;
end;
/

-- Avoid raising an exception

create or replace function insert_and_or_return_user3 (
   p_nm             in  my_users.nm%type,
   p_email_address  in  my_users.email_address%type
) return my_users.id%type
   authid definer
is
   cursor cur is
      select id
        from my_users
       where email_address = p_email_address;
   l_id my_users.id%type;
begin
   open cur;
   fetch cur into l_id;
   if cur%notfound then
      insert into my_users (
         nm,
         email_address
      ) values (
         p_nm,
         p_email_address
      ) return id into l_id;
   end if;
   close cur;
   return l_id;
end;
/

-- Use a merge with an empty update (if matched) clause

create or replace function insert_and_or_return_user4 (
   p_nm             in  my_users.nm%type,
   p_email_address  in  my_users.email_address%type
) return my_users.id%type
   authid definer
is
   l_id my_users.id%type;
begin
   merge into my_users u
   using (
            select nm,
                   email_address
              from my_users
             where email_address = p_email_address
         ) s
      on ( u.email_address = s.email_address )
    /* No update operation! when matched then */
    when not matched then
      insert (
         nm,
         email_address
      ) values (
         p_nm,
         p_email_address
      );
      
   /* Cannot use returning with merge so...*/
   select id
     into l_id
     from my_users
    where email_address = p_email_address;

   return l_id;   
end;
/

create or replace procedure test_variations (
   p_find_match  in  boolean,
   p_iterations  in  integer
) is
   l_id my_users.id%type;
   l_type           varchar2(100) :=
      case
         when p_find_match then
            'Match Found'
         else 'Match Not Found'
      end;
   l_nm             my_users.nm%type := 'Of No Importance';
   l_email_address  my_users.email_address%type := 'steven'
                                                  ||
      case
         when p_find_match then
            '500'
         else '100000'
      end
                                                  || '@insum.ca';
begin
   dbms_output.put_line(l_type
                        || ' with '
                        || p_iterations
                        || ' iterations');
   tmr.start_timing('query then insert if NDF');
   for indx in 1..p_iterations loop
      l_id :=
      insert_and_or_return_user1(
         p_nm             => l_nm,
         p_email_address  => l_email_address
      );
   end loop;
   tmr.show_elapsed;
   
   tmr.start_timing('insert then query if DVOI');
   for indx in 1..p_iterations loop
      l_id :=
      insert_and_or_return_user2(
         p_nm             => l_nm,
         p_email_address  => l_email_address
      );
   end loop;
   tmr.show_elapsed;
   
   tmr.start_timing('explicit cursor no exceptions');
   for indx in 1..p_iterations loop
      l_id :=
      insert_and_or_return_user3(
         p_nm             => l_nm,
         p_email_address  => l_email_address
      );
   end loop;
   tmr.show_elapsed;
   
   tmr.start_timing('merge no exceptions');
   for indx in 1..p_iterations loop
      l_id :=
      insert_and_or_return_user4(
         p_nm             => l_nm,
         p_email_address  => l_email_address
      );
   end loop;
   tmr.show_elapsed;
end;
/

begin 
test_variations (
   p_find_match  => true,
   p_iterations  => 100000
) ;
test_variations (
   p_find_match  => false,
   p_iterations  => 100000
) ;
end;
/