CREATE TABLE employees
AS
   SELECT * FROM hr.employees
/

/* Trying to bind a column name... */

CREATE OR REPLACE PROCEDURE set_to_10000 (col_in             IN VARCHAR2,
                                          department_id_in   IN PLS_INTEGER)
IS
   l_update   VARCHAR2 (1000) := 'UPDATE employees SET :colname = 10000 
        WHERE department_id = :dept';
BEGIN
   /* Basic protection against SQL injection */
   IF col_in NOT IN ('SALARY', 'COMMISSION_PCT')
   THEN
      RAISE_APPLICATION_ERROR (-20000,
                               'Unable to set value for ' || employee_id_in);
   END IF;

   EXECUTE IMMEDIATE l_update USING col_in, department_id_in;

   DBMS_OUTPUT.put_line ('Rows updated: ' || TO_CHAR (SQL%ROWCOUNT));
END;
/

BEGIN
   set_to_10000 ('salary', 50);
/* ORA-01747: invalid user.table.column, table.column, or column specification */
END;
/

/* Single bind variable */

CREATE OR REPLACE PROCEDURE set_to_10000 (col_in             IN VARCHAR2,
                                          department_id_in   IN PLS_INTEGER)
IS
   l_update   VARCHAR2 (1000)
      :=    'UPDATE employees SET '
         || col_in
         || ' = 10000 
        WHERE department_id = :dept';
BEGIN
   /* Basic protection against SQL injection */
   IF col_in NOT IN ('SALARY', 'COMMISSION_PCT')
   THEN
      RAISE_APPLICATION_ERROR (-20000,
                               'Unable to set value for ' || employee_id_in);
   END IF;

   EXECUTE IMMEDIATE l_update USING department_id_in;

   DBMS_OUTPUT.put_line ('Rows updated: ' || TO_CHAR (SQL%ROWCOUNT));
END;
/

DECLARE
   l_count   INTEGER;
BEGIN
   SELECT COUNT (*)
     INTO l_count
     FROM employees
    WHERE department_id = 50 AND salary = 10000;

   DBMS_OUTPUT.put_line ('Before ' || l_count);
   set_to_10000 ('salary', 50);

   SELECT COUNT (*)
     INTO l_count
     FROM employees
    WHERE department_id = 50 AND salary = 10000;

   DBMS_OUTPUT.put_line ('After ' || l_count);
   ROLLBACK;
END;
/

/* Multiple bind variables */

CREATE OR REPLACE PROCEDURE updnumval (col_in             IN VARCHAR2,
                                       department_id_in   IN PLS_INTEGER,
                                       val_in             IN NUMBER)
IS
   l_update   VARCHAR2 (1000)
      :=    'UPDATE employees SET '
         || col_in
         || ' = :val 
        WHERE department_id = :dept';
BEGIN
   EXECUTE IMMEDIATE l_update USING val_in, department_id_in;

   DBMS_OUTPUT.put_line ('Rows updated: ' || TO_CHAR (SQL%ROWCOUNT));
END;
/

DECLARE
   l_count   INTEGER;
BEGIN
   SELECT COUNT (*)
     INTO l_count
     FROM employees
    WHERE department_id = 50 AND salary = 10000;

   DBMS_OUTPUT.put_line ('Before ' || l_count);

   updnumval ('salary', 50, 10000);

   SELECT COUNT (*)
     INTO l_count
     FROM employees
    WHERE department_id = 50 AND salary = 10000;

   DBMS_OUTPUT.put_line ('After ' || l_count);
   ROLLBACK;
END;
/

/* Reference same bind variable more than once.

   Must supply value both times.
*/

CREATE OR REPLACE PROCEDURE updnumval (col_in             IN VARCHAR2,
                                       department_id_in   IN PLS_INTEGER,
                                       val_in             IN NUMBER)
IS
   l_update   VARCHAR2 (1000)
      :=    'UPDATE employees SET '
         || col_in
         || ' = :val 
        WHERE department_id = :dept and :val IS NOT NULL';
BEGIN
   /* Basic protection against SQL injection */
   IF col_in NOT IN ('SALARY', 'COMMISSION_PCT')
   THEN
      RAISE_APPLICATION_ERROR (-20000,
                               'Unable to set value for ' || employee_id_in);
   END IF;
   
   EXECUTE IMMEDIATE l_update USING val_in, department_id_in, val_in;

   DBMS_OUTPUT.put_line ('Rows updated: ' || TO_CHAR (SQL%ROWCOUNT));
END;
/

DECLARE
   l_count   INTEGER;
BEGIN
   SELECT COUNT (*)
     INTO l_count
     FROM employees
    WHERE department_id = 50 AND salary = 10000;

   DBMS_OUTPUT.put_line ('Before ' || l_count);

   updnumval ('salary', 50, 10000);

   SELECT COUNT (*)
     INTO l_count
     FROM employees
    WHERE department_id = 50 AND salary = 10000;

   DBMS_OUTPUT.put_line ('After ' || l_count);
   ROLLBACK;
END;
/

/* Single row changed */

CREATE OR REPLACE PROCEDURE set_one_row (col_in           IN VARCHAR2,
                                         employee_id_in   IN PLS_INTEGER)
IS
   l_update   VARCHAR2 (1000)
      :=    'UPDATE employees SET '
         || col_in
         || ' = 10000 
        WHERE employee_id = :empid
        RETURNING last_name INTO :the_name';
   l_name     VARCHAR2 (100);
BEGIN
   /* Basic protection against SQL injection */
   IF col_in NOT IN ('SALARY', 'COMMISSION_PCT')
   THEN
      RAISE_APPLICATION_ERROR (-20000,
                               'Unable to set value for ' || employee_id_in);
   END IF;

   EXECUTE IMMEDIATE l_update USING IN employee_id_in RETURNING INTO l_name;

   DBMS_OUTPUT.put_line ('Name updated: ' || l_name);
END;
/

BEGIN
   set_one_row ('salary', 138);
   ROLLBACK;
END;
/

/* Multiple rows changed */

CREATE OR REPLACE PROCEDURE set_multiple_rows (
   col_in             IN VARCHAR2,
   department_id_in   IN PLS_INTEGER)
IS
   l_update   VARCHAR2 (1000)
      :=    'UPDATE employees SET '
         || col_in
         || ' = 10000 
        WHERE department_id = :deptid
        RETURNING last_name INTO :the_name';
   l_name     VARCHAR2 (100);
BEGIN
   /* Basic protection against SQL injection */
   IF col_in NOT IN ('SALARY', 'COMMISSION_PCT')
   THEN
      RAISE_APPLICATION_ERROR (-20000,
                               'Unable to set value for ' || employee_id_in);
   END IF;

   EXECUTE IMMEDIATE l_update USING IN department_id_in RETURNING INTO l_name;

   DBMS_OUTPUT.put_line ('Name updated: ' || l_name);
END;
/

BEGIN
   set_multiple_rows ('salary', 50);
   /* ORA-01422: exact fetch returns more than requested number of rows */
   ROLLBACK;
END;
/

/* Need to BULK COLLECT INTO ... */

CREATE OR REPLACE PROCEDURE set_multiple_rows (
   col_in             IN VARCHAR2,
   department_id_in   IN PLS_INTEGER)
IS
   l_update   VARCHAR2 (1000)
      :=    'UPDATE employees SET '
         || col_in
         || ' = 10000 
        WHERE department_id = :deptid
        RETURNING last_name INTO :the_name';

   l_names    DBMS_SQL.varchar2a;
BEGIN
   /* Basic protection against SQL injection */
   IF col_in NOT IN ('SALARY', 'COMMISSION_PCT')
   THEN
      RAISE_APPLICATION_ERROR (-20000,
                               'Unable to set value for ' || employee_id_in);
   END IF;

   EXECUTE IMMEDIATE l_update
      USING IN department_id_in
      RETURNING BULK COLLECT INTO l_names;

   DBMS_OUTPUT.put_line ('Name updated: ' || l_names.COUNT);
END;
/

BEGIN
   set_multiple_rows ('salary', 50);
   ROLLBACK;
END;
/