/*
Demonstrates how to declare a schema-level nested table type, extend to make room for news and populate them,
use MULTISET operators to perform set level operations, and finally use the TABLE operator to fetch elements
from the nested table as though it were a relational table. Fun stuff!
*/

create or replace type list_of_names_t is
   table of varchar2(100);
/

grant execute on list_of_names_t to public
/

declare
   happyfamily    list_of_names_t := list_of_names_t();
   children       list_of_names_t := list_of_names_t();
   grandchildren  list_of_names_t := list_of_names_t();
   parents        list_of_names_t := list_of_names_t();
begin
   /* Can extend in "bulk" - 6 at once here */
   happyfamily.EXTEND(7);
   happyfamily(1)                         := 'Veva';
   happyfamily(2)                         := 'Chris';
   happyfamily(3)                         := 'Lauren';
   happyfamily(4)                         := 'Loey';
   happyfamily(5)                         := 'Eli';
   happyfamily(6)                         := 'Steven';
   happyfamily(7)                         := 'Juna';

   /* Individual extends. */
   children.EXTEND;
   children(children.LAST)                := 'Chris';
   children.EXTEND;
   children(children.LAST)                := 'Eli';
   children.EXTEND;
   children(children.LAST)                := 'Lauren';
   --
   grandchildren.EXTEND;
   grandchildren(grandchildren.LAST)      := 'Loey';
   grandchildren.EXTEND;
   grandchildren(grandchildren.LAST)      := 'Juna';

   /* Multiset operators on nested tables */
   parents                                 := ( happyfamily multiset except children ) multiset except grandchildren;
   for l_row in 1..parents.COUNT loop
      DBMS_OUTPUT.put_line(parents(l_row));
   end loop;

   happyfamily.delete;
   happyfamily.EXTEND(4);
   happyfamily(1)                         := 'Eli';
   happyfamily(2)                         := 'Steven';
   happyfamily(3)                         := 'Chris';
   happyfamily(4)                         := 'Veva';

   /* Use TABLE operator to apply SQL operations to
      a PL/SQL nested table */

   for rec in (
      select COLUMN_VALUE family_name
        from table ( happyfamily )
       order by family_name
   ) loop
      DBMS_OUTPUT.put_line(rec.family_name);
   end loop;
end;
/

/* Using a nested table as a column type */

drop table family
/

CREATE OR REPLACE TYPE parent_names_t
   IS TABLE OF VARCHAR2 (100);
/

CREATE OR REPLACE TYPE child_names_t
   IS TABLE OF VARCHAR2 (100);
/

CREATE TABLE family
(
   surname          VARCHAR2 (1000)
 ,  parent_names     parent_names_t
 ,  children_names   child_names_t
)
NESTED TABLE children_names
   STORE AS parent_names_tbl -- (TABLESPACE USERS)
NESTED TABLE parent_names
   STORE AS children_names_tbl -- (TABLESPACE TEMP)
/

DECLARE
   parents    parent_names_t
                 := parent_names_t ('Steven', 'Veva');children child_names_t : = child_names_t ( 'Chris' , 'Eli' ) ; begin insert into family ( surname , parent_names , children_names ) values ( 'Feuerstein' , parents , children ) ; commit ; end ;
/ select * from family
/