drop type ticker_ot force;
drop type tickers_nt force;
drop type stock_ot force;
drop type stocks_nt force;
drop table stocks;
drop table tickers;

create table stocks (
   ticker       varchar2(20),
   trade_date   date,
   open_price   number,
   close_price  number
)
/

/* Load up 10000 rows - when running in your own database, you might want to
   use a higher volume of data here, to see a more dramatic difference in the
   elapsed time and memory utilization 
*/

insert into stocks
   select 'STK' || LEVEL,
          SYSDATE,
          LEVEL,
          LEVEL + 15
     from DUAL connect by
      LEVEL <= 1000000
/

create table tickers (
   ticker     varchar2(20),
   pricedate  date,
   pricetype  varchar2(1),
   price      number
)
/

create type ticker_ot as object (
   ticker     varchar2(20),
   pricedate  date,
   pricetype  varchar2(1),
   price      number
)
/

create type tickers_nt as
   table of ticker_ot;
/

create or replace package stock_mgr authid DEFINER is
   type stocks_rc is ref cursor return stocks%ROWTYPE;
end stock_mgr;
/

create or replace function doubled_nopl (
   rows_in stock_mgr.stocks_rc
) return tickers_nt
   authid DEFINER
is
   type stocks_aat is
      table of stocks%ROWTYPE index by pls_integer;
   l_stocks   stocks_aat;
   l_doubled  tickers_nt := tickers_nt();
begin
   loop
      fetch rows_in
        bulk collect into l_stocks limit 100;
      exit when l_stocks.COUNT = 0;
      for l_row in 1..l_stocks.COUNT loop
         l_doubled.EXTEND;
         l_doubled(l_doubled.LAST)       
            := ticker_ot(l_stocks(l_row).ticker, l_stocks(l_row).trade_date, 'O', l_stocks(l_row).open_price);

         l_doubled.EXTEND;
         l_doubled(l_doubled.LAST)       
            := ticker_ot(l_stocks(l_row).ticker, l_stocks(l_row).trade_date, 'C', l_stocks(l_row).close_price);
      end loop;
   end loop;

   return l_doubled;
end;
/

create or replace function doubled_pl (
   rows_in stock_mgr.stocks_rc
) return tickers_nt
   pipelined
   authid DEFINER
is
   type stocks_aat is
      table of stocks%ROWTYPE index by pls_integer;
   l_stocks stocks_aat;
begin
   loop
      fetch rows_in
        bulk collect into l_stocks limit 100;
      exit when l_stocks.COUNT = 0;
      for l_row in 1..l_stocks.COUNT 
      loop
         pipe row ( ticker_ot(l_stocks(l_row).ticker, 
                              l_stocks(l_row).trade_date, 
                              'O', 
                              l_stocks(l_row).open_price) );

         pipe row ( ticker_ot(l_stocks(l_row).ticker, 
                              l_stocks(l_row).trade_date, 
                              'C', 
                              l_stocks(l_row).close_price) );
      end loop;
   end loop;

   return;
end;
/

create or replace package utils is
   procedure initialize (
      context_in in varchar2
   );

   procedure show_results (
      message_in in varchar2 := null
   );
end;
/

create or replace package body utils is
   last_timing  integer := null;
   last_pga     integer := null;

   function pga_consumed return number as
      l_pga number;
   begin
      select st.VALUE
        into l_pga
        from v$mystat st,
             v$statname sn
       where st.statistic# = sn.statistic#
         and sn.name = 'session pga memory';

      return l_pga;
   end;

   procedure initialize (
      context_in in varchar2
   ) is
   begin
      delete from tickers;
      commit;
      DBMS_OUTPUT.put_line(context_in);
      last_timing  := DBMS_UTILITY.get_time;
      last_pga     := pga_consumed;
   end;

   procedure show_results (
      message_in in varchar2 := null
   ) is
      l_count integer;
   begin
      select count(*)
        into l_count
        from tickers;

      DBMS_OUTPUT.put_line('Ticker row count: ' || l_count);
      DBMS_OUTPUT.put_line(
         '"'
         || message_in
         || '" completed in: '
         || TO_CHAR(
            DBMS_UTILITY.get_time - last_timing
         )
         || ' centisecs; pga at: '
         || TO_CHAR(
            pga_consumed() - last_pga
         )
         || ' bytes'
      );
   end;
end;
/

begin
   utils.initialize('Pipelined');
   insert into tickers
      select *
        from table ( doubled_pl(cursor(
                      select *
                        from stocks
                   )) )
       where rownum < 10;

   utils.show_results('First 9 rows');
   utils.initialize('Not Pipelined');
   insert into tickers
      select *
        from table ( doubled_nopl(cursor(
                      select *
                        from stocks
                   )) )
       where rownum < 10;

   utils.show_results('First 9 rows');
end;
/

/* Penalty using PTF when not needed? Not much....

Pipelined
1980000
Ticker row count: 0
"10000 invocations" completed in: 753 centisecs; pga at: 65536 bytes
Not Pipelined
3960000
Ticker row count: 0
"10000 invocations" completed in: 1004 centisecs; pga at: 56098816 bytes

*/
declare
   l_iterations integer := 10000;
   l_count integer := 0;
begin
   utils.initialize('Pipelined');
   for indx in 1..l_iterations loop
      for rec in (
         select *
           from table ( doubled_pl(cursor(
                         select *
                           from stocks
                          where rownum < 100
                      )) )
      ) loop
         l_count := l_count + 1;
      end loop;
   end loop;
      dbms_output.put_line(l_count);

   utils.show_results(l_iterations|| ' invocations');
   utils.initialize('Not Pipelined');
   for indx in 1..l_iterations loop
      for rec in (
         select *
           from table ( doubled_nopl(cursor(
                         select *
                           from stocks
                          where rownum < 100
                      )) )
      ) loop
         l_count := l_count + 1;
      end loop;
   end loop;
      dbms_output.put_line(l_count);

   utils.show_results(l_iterations|| ' invocations');
end;
/