create or replace type list_of_names_t is
   table of varchar2(100);
/

declare
   big_brained_species list_of_names_t := list_of_names_t(
      'Human', -- neocortex
      'Bonobo', -- neocortex
      'Gorilla', -- neocortex
      'Octopus', -- NINE BRAINS!
      'Raven', -- pallium
      'Whale', -- neocortex
      'Dolphin' -- neocortex
   );
begin
   for rec in (
      select column_value species_name
        from table ( big_brained_species )
       order by species_name
   ) loop
      dbms_output.put_line(rec.species_name);
   end loop;
end;
/

-- where clauses

declare
   big_brained_species list_of_names_t := list_of_names_t(
      'Human',
      'Bonobo',
      'Gorilla',
      'Octopus',
      'Raven',
      'Whale',
      'Dolphin'
   );
begin
   for rec in (
      select column_value species_name
        from table (big_brained_species)
       where column_value like '%o%'
   ) loop
      dbms_output.put_line(rec.species_name);
   end loop;
end;
/

-- joins: combine persistent data with in-session data

declare
   big_brained_species list_of_names_t := list_of_names_t(
      'Human',
      'Bonobo',
      'Gorilla',
      'Octopus',
      'Raven',
      'Whale',
      'Dolphin'
   );
begin
   for rec in (
      select column_value || '-' || e.last_name name
        from table (big_brained_species) bbs, employees e
       where substr (bbs.column_value, 1, 1) = substr (e.last_name, 1, 1)
   ) loop
      dbms_output.put_line(rec.name);
   end loop;
end;
/

-- what about a collection defined in a package?

create or replace package brainiacs is
   big_brained_species list_of_names_t := list_of_names_t(
      'Human',
      'Bonobo',
      'Gorilla',
      'Octopus',
      'Raven',
      'Whale',
      'Dolphin'
   );
end;
/

-- PLS-221: 'BIG_BRAINED_SPECIES' is not a procedure or is undefined
-- Same old same old: cannot yet directly reference PL/SQL vars in SQL-outside-of-PLSQL

select column_value species_name
  from table ( brainiacs.big_brained_species )
 order by species_name
/

-- No problem *inside* PLSQL

begin
   for rec in (
      select column_value species_name
        from table ( brainiacs.big_brained_species )
       order by species_name
   ) loop
      dbms_output.put_line(rec.species_name);
   end loop;
end;
/

/* Now let's put that code inside a function */

create or replace function big_brained_species return list_of_names_t is
begin
   return list_of_names_t(
      'Human',
      'Bonobo',
      'Gorilla',
      'Octopus',
      'Raven',
      'Whale',
      'Dolphin'
   );
end;
/

select column_value species_name
  from table ( big_brained_species )
 order by species_name
/

-- ORA-04044: procedure, function, package, or type is not allowed here

select column_value species_name
  from big_brained_species
 order by species_name
/

-- No problem

select column_value species_name
  from big_brained_species ()
 order by species_name
/

-- Non-scalar collections (based on object types or record types)
-- Each attribute or field name -> column name

drop package brainiacs
/

create or replace package brainiacs 
is
   type species_rt is record (name varchar2(100), average_size varchar2(100));
   type species_aat is table of species_rt index by pls_integer;
   
   big_brained_species species_aat := species_aat(
      -10000 => species_rt ('Human', 'Medium'),
      1000 => species_rt ('Bonobo', 'Medium'),
      2000 => species_rt ('Gorilla', 'Large'),
      3000 => species_rt ('Octopus', 'Small'),
      4000 => species_rt ('Raven', 'Very Small'),
      5000 => species_rt ('Whale', 'Very Large'),
      6000 => species_rt ('Dolphin', 'Medium')
   );
end;
/

begin
   for rec in (
      select name, average_size
        from table ( brainiacs.big_brained_species )
       order by name desc
   ) loop
      dbms_output.put_line(rec.average_size || ' ' || rec.name);
   end loop;
end;
/

create or replace package brainiacs 
is
   type species_rt is record (name varchar2(100), average_size varchar2(100));
   type species_aat is table of species_rt index by pls_integer;
   
   function big_brained_species (p_like in varchar2 default '%') return species_aat;
end;
/

create or replace package body brainiacs 
is
   function big_brained_species (
      p_like in varchar2 default '%'
   ) return species_aat is
   begin
      return species_aat(
         -10000   => species_rt('Human', 'Medium'),
         1000     => species_rt('Bonobo', 'Medium'),
         2000     => species_rt('Gorilla', 'Large'),
         3000     => species_rt('Octopus', 'Small'),
         4000     => species_rt('Raven', 'Very Small'),
         5000     => species_rt('Whale', 'Very Large'),
         6000     => species_rt('Dolphin', 'Medium')
      );
   end;
end;   
/

-- But strangely enough cannot access in SQL directly inside PL/SQL

begin
   for rec in (
      select name, average_size
        from table ( brainiacs.big_brained_species() )
       order by name desc
   ) loop
      dbms_output.put_line(rec.average_size || ' ' || rec.name);
   end loop;
end;    
/

-- Call inside PL/SQL, assign to local variable

declare
   l_species brainiacs.species_aat := brainiacs.big_brained_species();
begin
   for rec in (
      select name, average_size
        from table ( l_species )
       order by name desc
   ) loop
      dbms_output.put_line(rec.average_size || ' ' || rec.name);
   end loop;
end;    
/

-- Let's try with object types

create or replace type species_ot is object
   (name varchar2(100), average_size varchar2(100))
/

create or replace type species_nt as table of species_ot
/

create or replace package brainiacs 
is
   function big_brained_species (p_like in varchar2 default '%') return species_nt;
end;
/

create or replace package body brainiacs 
is
   function big_brained_species (
      p_like in varchar2 default '%'
   ) return species_nt is
   begin
      return species_nt(
         species_ot('Human', 'Medium'),
         species_ot('Bonobo', 'Medium'),
         species_ot('Gorilla', 'Large'),
         species_ot('Octopus', 'Small'),
         species_ot('Raven', 'Very Small'),
         species_ot('Whale', 'Very Large'),
         species_ot('Dolphin', 'Medium')
      );
   end;
end;   
/

-- Now, no problem.

begin
   for rec in (
      select name, average_size
        from table ( brainiacs.big_brained_species() )
       order by name desc
   ) loop
      dbms_output.put_line(rec.average_size || ' ' || rec.name);
   end loop;
end;    
/